#include "virtual-keyboard-widgets.h"

/**
 * @brief DiyaVkbButton: virtual keyboard button
 */

struct _DiyaVkbButton
{
    GtkWidget super;
    GtkWidget *labels[3];
    DiyaVkbKey *key; /*owned by the virtual keyboard*/
    gulong mod_sig_id;
};

G_DEFINE_FINAL_TYPE(DiyaVkbButton, diya_vkb_button, GTK_TYPE_WIDGET);

static void diya_vkb_button_dispose(GObject *object)
{
    DiyaVkbButton *self = DIYA_VKB_BUTTON(object);
    g_debug("diya_vkb_button_dispose");
    for (int i = 0; i < 3; i++)
    {
        g_clear_pointer(&self->labels[i], gtk_widget_unparent);
    }
    //  g_clear_pointer (&demo->label, gtk_widget_unparent);
    G_OBJECT_CLASS(diya_vkb_button_parent_class)->dispose(object);
}

static void diya_vkb_button_event(GtkGestureClick *gesture, int n_press, double x, double y, gpointer data)
{
    (void) n_press;
    (void) x;
    (void) y;
    const gchar* evt_name = (gchar*) data;
    GtkEventController *controller = GTK_EVENT_CONTROLLER(gesture);
    DiyaVkbButton *self = DIYA_VKB_BUTTON(gtk_event_controller_get_widget(controller));
    g_signal_emit_by_name(self, evt_name);
}

static void diya_vkb_button_init(DiyaVkbButton *self)
{
    gchar class_name[32]; /*diya-vkb-btn-level-%d */
    GtkLayoutManager *grid_layout;
    self->key = NULL;
    grid_layout = gtk_widget_get_layout_manager(GTK_WIDGET(self));
    // gtk_orientable_set_orientation(GTK_ORIENTABLE(box_layout), GTK_ORIENTATION_VERTICAL);
    //  gtk_box_layout_set_spacing (GTK_BOX_LAYOUT (box_layout), 3);
    gtk_widget_set_can_focus(GTK_WIDGET(self), false);
    // init labels
    for (int i = 0; i < 3; i++)
    {
        self->labels[i] = gtk_label_new(NULL);
        g_snprintf(class_name, sizeof(class_name), "shift-level-%d", i + 1);
        gtk_widget_add_css_class(self->labels[i], class_name);
        gtk_widget_set_hexpand(self->labels[i], true);

        gtk_widget_set_parent(self->labels[i], GTK_WIDGET(self));

        GtkGridLayoutChild *grid_child = GTK_GRID_LAYOUT_CHILD(gtk_layout_manager_get_layout_child(grid_layout, self->labels[i]));
        switch (i)
        {
        case 0:
            /* code */
            gtk_grid_layout_child_set_row(grid_child, 0);
            gtk_grid_layout_child_set_column(grid_child, 0);
            gtk_grid_layout_child_set_column_span(grid_child, 2);
            break;
        case 1:
            gtk_grid_layout_child_set_row(grid_child, 1);
            gtk_grid_layout_child_set_column(grid_child, 0);
            gtk_widget_set_halign(self->labels[i], GTK_ALIGN_START);
            break;
        case 2:
            gtk_grid_layout_child_set_row(grid_child, 1);
            gtk_grid_layout_child_set_column(grid_child, 1);
            gtk_widget_set_halign(self->labels[i], GTK_ALIGN_END);
            break;
        default:
            break;
        }
    }
    // event handle
    GtkGesture *controller = gtk_gesture_click_new();
    g_signal_connect(controller, "pressed", G_CALLBACK(diya_vkb_button_event), DIYA_SIGNAL_VKB_KEY_PRESSED);
    g_signal_connect(controller, "released", G_CALLBACK(diya_vkb_button_event), DIYA_SIGNAL_VKB_KEY_RELEASED);
    gtk_widget_add_controller(GTK_WIDGET(self), GTK_EVENT_CONTROLLER(controller));
}

static void diya_vkb_button_class_init(DiyaVkbButtonClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    gobject_class->dispose = diya_vkb_button_dispose;
    gtk_widget_class_set_css_name(GTK_WIDGET_CLASS(class), "diya-vkb-button");
    gtk_widget_class_set_layout_manager_type(GTK_WIDGET_CLASS(class), GTK_TYPE_GRID_LAYOUT);
}

void diya_vkb_button_set_key(DiyaVkbButton* btn, DiyaVkbKey *key)
{
    btn->key = key;
    for (uint32_t i = 0; i < 3; i++)
    {
        if(key)
        {
            const char *cap = diya_vkb_key_get_keycap(key, i);
            if (cap)
            {
                if (i == 1 && diya_vkb_key_is_alphabet(key))
                {
                    continue;
                }
                gtk_label_set_text(GTK_LABEL(btn->labels[i]), cap);
            }
        }
        else
        {
            gtk_label_set_text(GTK_LABEL(btn->labels[i]), NULL);
        }
    }
}

DiyaVkbButton *diya_vkb_button_new(DiyaVkbKey *key)
{
    DiyaVkbButton *btn = DIYA_VKB_BUTTON(g_object_new(DIYA_TYPE_VKB_BUTTON, NULL));
    if(key)
    {
        diya_vkb_button_set_key(btn, key);
    }
    btn->mod_sig_id = 0;
    return btn;
}

/**
 * @brief DiyaVirtualKeyboardWidget: Virtual keyboard GTK widget
 *
 */

#define VKB_N_CELLS 145
#define VKB_ROW_NUM(i) (i / 29)
#define VKB_COL_NUM(i) (i % 29)

static uint32_t g_vkb_layout[] = {
    49, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22,
    23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 51,
    37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 36,
    50, 94, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 111, 66,
    9, 206, 64, 65, 0, 108, 113, 116, 114};

struct _DiyaVirtualKeyboardWidget
{
    GtkWidget super;
    /**
     * row 0: 00 - 13
     * row 1: 14 - 27
     * row 2: 28 - 40
     * row 3: 41 - 53
     * row 4: 54 - 62
     */
    GList *buttons;
    DiyaVirtualKeyboard *vkb; /*owned by the widget itself*/
};

G_DEFINE_FINAL_TYPE(DiyaVirtualKeyboardWidget, diya_virtual_keyboard_widget, GTK_TYPE_WIDGET);

static void diya_virtual_keyboard_widget_dispose(GObject *object)
{
    DiyaVirtualKeyboardWidget *self = DIYA_VIRTUAL_KEYBOARD_WIDGET(object);
    g_debug("diya_virtual_keyboard_widget_dispose");
    if (self->buttons)
    {
        GList *it = NULL;
        for (it = self->buttons; it; it = it->next)
        {
            if (it->data)
            {
                g_clear_pointer(&it->data, gtk_widget_unparent);
            }
        }
    }
    /* if(self->vkb)
    {
        g_object_unref(self->vkb);
    } */
    //  g_clear_pointer (&demo->label, gtk_widget_unparent);
    G_OBJECT_CLASS(diya_virtual_keyboard_widget_parent_class)->dispose(object);
}

static void diya_virtual_keyboard_widget_init(DiyaVirtualKeyboardWidget *self)
{
    // gtk_widget_set_size_request(GTK_WIDGET(self), 600, 400);
    g_debug("diya_virtual_keyboard_widget_init: Create new widget");
    GtkLayoutManager *grid_layout = gtk_widget_get_layout_manager(GTK_WIDGET(self));
    // gtk_orientable_set_orientation(GTK_ORIENTABLE(grid_layout), GTK_ORIENTATION_VERTICAL);
    gtk_grid_layout_set_row_spacing(GTK_GRID_LAYOUT(grid_layout), 2);
    gtk_grid_layout_set_column_spacing(GTK_GRID_LAYOUT(grid_layout), 2);
    // gtk_grid_layout_set_baseline_row(GTK_GRID_LAYOUT(grid_layout), 4);
    //  init buttons
    self->vkb = NULL;
}

static void diya_virtual_keyboard_widget_class_init(DiyaVirtualKeyboardWidgetClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    gobject_class->dispose = diya_virtual_keyboard_widget_dispose;
    gtk_widget_class_set_css_name(GTK_WIDGET_CLASS(class), "diya-vkb");
    gtk_widget_class_set_layout_manager_type(GTK_WIDGET_CLASS(class), GTK_TYPE_GRID_LAYOUT);
    g_signal_new(DIYA_SIGNAL_VKB_KEY_PRESSED,
        DIYA_TYPE_VKB_BUTTON,
        G_SIGNAL_DETAILED |
            G_SIGNAL_ACTION |
            G_SIGNAL_RUN_FIRST,
        0,
        NULL,
        NULL,
        NULL,
        G_TYPE_NONE,
        0,
        NULL);
    g_signal_new(DIYA_SIGNAL_VKB_KEY_RELEASED,
        DIYA_TYPE_VKB_BUTTON,
        G_SIGNAL_DETAILED |
            G_SIGNAL_ACTION |
            G_SIGNAL_RUN_FIRST,
        0,
        NULL,
        NULL,
        NULL,
        G_TYPE_NONE,
        0,
        NULL);
}

static void diya_vkb_widget_on_key_pressed(DiyaVkbButton* button, DiyaVirtualKeyboardWidget* self)
{
    gtk_widget_add_css_class(GTK_WIDGET(button), "active");
    diya_virtual_keyboard_send_key(self->vkb, button->key,VKB_KEY_STATE_PRESSED);
}

static void diya_vkb_widget_on_key_released(DiyaVkbButton* button, DiyaVirtualKeyboardWidget* self)
{
    if(!diya_vkb_key_is_modifier(button->key))
    {
        gtk_widget_remove_css_class(GTK_WIDGET(button), "active");
    }
    diya_virtual_keyboard_send_key(self->vkb, button->key,VKB_KEY_STATE_RELEASED);
}

static void on_kb_modifier_changed(DiyaVirtualKeyboard* kb, DiyaVkbButton* button)
{
    if(!diya_virtual_keyboard_key_modifier_is_active(kb, button->key))
    {
        gtk_widget_remove_css_class(GTK_WIDGET(button), "active");
    }
}

void diya_virtual_keyboard_widget_set_keyboard(DiyaVirtualKeyboardWidget* self, DiyaVirtualKeyboard* keyboard)
{
    self->vkb = keyboard;
    int index = 0;
    if (self->buttons)
    {
        GList *it = NULL;
        for (it = self->buttons; it; it = it->next)
        {
            if (it->data)
            {
                DiyaVkbButton* button = it->data;
                DiyaVkbKey *key = NULL;
                if(button->mod_sig_id > 0)
                {
                    g_signal_handler_disconnect(self->vkb,button->mod_sig_id);
                    button->mod_sig_id = 0;
                }
                if(keyboard)
                {
                    key = diya_virtual_keyboard_get_key(self->vkb, g_vkb_layout[index]);
                    g_debug("set button key: %s", diya_object_to_string(key));
                    if(diya_vkb_key_is_modifier(key))
                    {
                        button->mod_sig_id = g_signal_connect(self->vkb, DIYA_SIGNAL_VKB_MODIFIER_CHANGED, G_CALLBACK(on_kb_modifier_changed), button);
                    }
                }
                diya_vkb_button_set_key(button, key);
                index++;
            }
        }
    }
}

GtkWidget *diya_virtual_keyboard_widget_new(DiyaVirtualKeyboard* keyboard)
{
    DiyaVirtualKeyboardWidget *self = g_object_new(DIYA_TYPE_VIRTUAL_KEYBOARD_WIDGET, NULL);
    GtkLayoutManager *grid_layout = gtk_widget_get_layout_manager(GTK_WIDGET(self));
    self->vkb = keyboard;
    if(self->vkb == NULL)
    {
        return NULL;
    }
    for (int i = 0; i < VKB_N_CELLS; i++)
    {
        gboolean skip = false;
        int row = VKB_ROW_NUM(i);
        int col = VKB_COL_NUM(i);
        switch (row)
        {
        case 0:
            skip = (col % 2 != 0) || (col > 26);
            break;
        case 1:
            skip = (col == 1) || (col == 2) || (col != 0 && col % 2 == 0);
            break;
        case 2:
            skip = (col == 1) || (col == 2) || (col != 0 && col % 2 == 0) || (col > 25);
            break;
        case 3:
            skip = (col > 0 && col < 3) || (col != 0 && col % 2 == 0);
            break;
        case 4:
            skip = (col > 0 && col < 3) || (col == 4) || (col == 6) || (col > 7 && col < 19) || (col > 19 && col % 2 == 0);
            break;
        default:
            break;
        }

        if (skip)
        {
            continue;
        }
        
        GtkWidget *button = GTK_WIDGET(diya_vkb_button_new(NULL));
        self->buttons = g_list_append(self->buttons, button);
        gtk_widget_set_parent(button, GTK_WIDGET(self));
        gtk_widget_set_hexpand(button, true);
        GtkGridLayoutChild *grid_child = GTK_GRID_LAYOUT_CHILD(gtk_layout_manager_get_layout_child(grid_layout, button));

        g_signal_connect(button, "pressed", G_CALLBACK(diya_vkb_widget_on_key_pressed),self);
        g_signal_connect(button, "released", G_CALLBACK(diya_vkb_widget_on_key_released),self);

        gtk_grid_layout_child_set_row(grid_child, row);
        gtk_grid_layout_child_set_column(grid_child, col);

        switch (i)
        {
        case 26:
        case 29:
        case 58:
        case 87:
        case 116:
            gtk_grid_layout_child_set_column_span(grid_child, 3);
            break;

        case 83:
            gtk_grid_layout_child_set_column_span(grid_child, 4);
            break;
       /*  case 87:
            gtk_grid_layout_child_set_column_span(grid_child, 5); */
            break;
        case 123:
            gtk_grid_layout_child_set_column_span(grid_child, 12);
            break;

        default:
            if (i > 0)
                gtk_grid_layout_child_set_column_span(grid_child, 2);
            break;
        }
    }
    if(keyboard)
    {
        diya_virtual_keyboard_widget_set_keyboard(self, keyboard);
    }
    return GTK_WIDGET(self);
}
